﻿//-----------------------------------------------------------------------------
//
// Title:			ds30 Loader
//
// Copyright:		Copyright © 08-10, Mikael Gustafsson
//
// Version:			1.3.3 january 2010
//
// Link:			http://mrmackey.no-ip.org/elektronik/ds30loader/
//
// History:			1.3.3 New feature: auto baudrate detection
//                        New feature: tx echo verification
//                        New feature: PIC16F support
//                        Improvement: FindLoader() discards all leading null bytes
//                        Improvement: debug output
//                  1.3.2 Bugfix: serious bug that may cause overwrite of bootloader when using custom bootlooader
//                  1.3.1 Major code rework for WriteFlash() and ParseHex() to allow for easier maintenance
//                        Improved hex file validation
//                        Improved debug output
//                        Fixed broken dtr activation 
//                        Added activations in FindLoader()
//                        If row and page 0 is not found in hex file, goto at 0x00 is not written
//                        Now reports if unknown data is found in the hex file
//                  1.3.0 Added possibility to use custom bootloader placement and size
//                  1.2.5 Fixed broken dtr activation
//                        Improved pic18 hex parse
//                  1.2.4 Fixed dtr and rts reset, probably not working previously
//                        Improved info to user
//                        Minor non-functional code improvements in ParseHex()*.*
//                        Dtr and rts are brought low when port is closed
//                  1.2.3 Input buffer is cleared when resetting
//                        New reset method: command sequence
//                  1.2.2 More improvements in FindLoader() to reduce time from hello to download
//                        Added public versionstring
//                        Alot of non-functional code improvements
//                        PIC18FJ: fixed bug in WriteFlash()
//                        PIC18FJ: fixed incorrect flashsize for most devices
//                        PIC18FJ: fixed bootloader placement to 2nd last page, configs are stored in last page
//                  1.2.1 Added support for PIC18F 1.5 version
//                        Minor improvements in FindLoader()
//                  1.2.0 Added a settings class that's to be passed to Download()
//                        Reset device options added
//                        Activate device options added
//                        Added debug output for all devicefamilies/firmware versions
//                        Improvements in FindLoader()
//                        Fixed Eeprom and config write for dsPIC30F firmware >= 2.0
//                  1.1.1 Fixed dsPIC30F3011 device id, from 3 to 13
//                  1.1.0 Major code rework making it more independent and easier to integrate to 3rd party gui
//                  1.0.4 Some improvements in clsSerialPort
//                        GUI and actual bootloader code is now separated
//                        clsSerialport and debugtext moved to separate helper library
//                  1.0.3 Added recognition of new fw versions
//                  1.0.2 Added support for PIC18FJ
//                        Added debugmode
//                  1.0.1 Fixed incorrect bootloder size for PIC18 with rowsize <> 16 words
//                  1.0.0 Added support for PIC24H
//                        Split PIC24F to PIC24F and PIC24FJ
//                        Added support for PIC18F
//                        Fixed som copy/paste errors
//                  0.9.9 Fixed configs not found in dsPIC33
//                        Fixed dsPIC33F config write, byte instead of word
//                  0.9.9 Added support for PIC24F
//                        Fixed incorrect addresscheck in ParseHex()
//                  0.9.8 Eeprom buffer is now properly sized
//						  Optimized ParseHex()
//						  Added support for dsPIC33F
//					0.9.7 Added support for Eeprom programming
//						  Better error handling
//						  Added chek for config count
//					0.9.6 Fixed checksum calc for configs
//                  0.9.5 Fixed broken support for dsPIC30F3013
//                        Added support for 1010, 2020, 2023, 5016 and 6015
//					0.9.4 Fixed a bug that caused invalid goto on devices with flash size > ??
//					0.9.3 Fixed a bug in ParseHex()
//					0.9.2 Support for new 0.9.2 algoritm where only 96bytes are sent for each row instead of 128, ~25% faster
//					0.9.0 Initial release
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//    This file is part of ds30 Loader.
//
//    ds30 Loader is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation.
//
//    ds30 Loader is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ds30 Loader.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------  

using System;
using GHelper;
using System.Threading;
using System.Windows.Forms;

namespace ds30Loader
{
    //-------------------------------------------------------------------------
	// Class: clsds30Loader
	//-------------------------------------------------------------------------
	static public class clsds30Loader
	{
		//---------------------------------------------------------------------
		// Size suffixes
		//---------------------------------------------------------------------
		// R - rows
		// W - words
		// P - program counter units
		// B3 - bytes				
		// B4 - bytes		

			
		//---------------------------------------------------------------------
		// Variables 
		//---------------------------------------------------------------------        
		static private bool bAbort = false;			//external abortsignal
        static private bool bDebugmode = false;


		//---------------------------------------------------------------------
		// Test-mode
		//--------------------------------------------------------------------- 
        static public bool testMode { get; set; }
        static public int testDeviceID { get; set; }
        static public Version testFwVer { get; set; }
        static public char testResponse { get; set; }


		//---------------------------------------------------------------------
		// Constants
		//---------------------------------------------------------------------        
        public const string strVersion = "1.3.3";

        public const byte cHello = 0xC1;
		public const char cOK = 'K';
		public const char cChecksumErr = 'N';
        public const char cVerifyErr = 'V';
        public const char cBlProtTrip = 'P';
		
		private const byte cCmdErasePage = 1;   //bit 0
		private const byte cCmdWriteRow = 2;    //bit 1
        private const byte cCmdWriteEEWord = 4; //bit 2
		private const byte cCmdWriteConfig = 8; //bit 3
        private const byte cCmdRead = 16;       //bit 4


		// 
		public delegate void DownloadingDelegate( object sender, clsDownloadingEventArgs e );
		static public event DownloadingDelegate Downloading;		


        //---------------------------------------------------------------------
		// Constructor()
		//---------------------------------------------------------------------
		static clsds30Loader()
        {
            testMode = false;
		}//Constructor()

		        
		//---------------------------------------------------------------------
		// Event: Downloading
		//---------------------------------------------------------------------
		static internal void OnDownloading( clsDownloadingEventArgs e )
		{
			// Event will be null if no client has hooked up a delegate to the event
			if ( Downloading != null ) {
				Downloading( null, e );
			}
		}//Event: FileValidated
		

 		//---------------------------------------------------------------------
		// Property: debugMode
		//---------------------------------------------------------------------
		static public bool debugMode
		{
			get {
				return bDebugmode;
			} set {
                 bDebugmode = value;
            }
        }//Property: family  	
	

		//---------------------------------------------------------------------
		// Abort()
		//---------------------------------------------------------------------
		static public void Abort()
		{
			bAbort = true;
		}// Abort()       

        
		//---------------------------------------------------------------------
		// Download()
		//---------------------------------------------------------------------
		static public void Download ( clsDevice pobjDevice, clsSerialPort pobjPort, clsHex pobjHex, clsDownloadSettings pobjSettings, int iTabLevel, ref bool pbResult )
		{
            //
            bool bWriteSuccess = false;
            bool bParseResult = false;

            pbResult = false;
            bAbort = false;
            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.started, "Initiating download...", iTabLevel++) );
            
            // Test-mode
            if ( testMode == true ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.warning, "Simulations mode is active", iTabLevel) );
            }


			//--------------------------------------------------------------------------
			// Reset device
			//--------------------------------------------------------------------------
            bool bResetResult = false;
            clsds30Loader.ResetDevice( pobjPort, pobjSettings, iTabLevel, ref bResetResult );
            if ( bResetResult == false ) {
                return;
            }

    			
		    //--------------------------------------------------------------------------
		    // Open port
		    //--------------------------------------------------------------------------
		    bool bOpenPortResult = false;
            PortOpen( pobjPort, pobjSettings, true, ref bOpenPortResult, iTabLevel );
            if ( bOpenPortResult == false ) {
                return;
            }


            //--------------------------------------------------------------------------
			// Find loader
			//--------------------------------------------------------------------------
            bool bFindLoaderResult = false;
            Version verFw = null;
            FindLoader( pobjDevice, pobjPort, pobjSettings, ref verFw, iTabLevel, ref bFindLoaderResult );
            if ( bFindLoaderResult == false ) {
				PortClose( pobjPort, iTabLevel );
                return;
            }
            string strFwVer = verFw.ToString();
            		    

			//--------------------------------------------------------------------------
			// Init progressbar
			//--------------------------------------------------------------------------
            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progressStarted, "", 0) );


			//--------------------------------------------------------------------------
			// Determine properties of the bootloader found
			//--------------------------------------------------------------------------
            int iBlSizeP;
            int iBlPlacementP;

            if ( pobjSettings.customBl == false ) {
                // Placement
                {
                    bool bGetBLPlacementResult = false;
                    iBlPlacementP = GetDefaultBlPlacementP( pobjDevice, verFw, ref bGetBLPlacementResult );
                    if ( bGetBLPlacementResult == false ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                        PortClose( pobjPort, iTabLevel );
                        return;
                    }
                    if ( debugMode == true ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Bootloader placement: " + iBlPlacementP.ToString(), iTabLevel) );
                    }
                }

                // Size
                {
                    bool bGetBLSizeResult = false;
                    iBlSizeP = GetDefaultBLSizeP( pobjDevice, verFw, ref bGetBLSizeResult );
                    if ( bGetBLSizeResult == false ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                        PortClose( pobjPort, iTabLevel );
                        return;
                    }
                    if ( debugMode == true ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Bootloader size: " + iBlSizeP.ToString(), iTabLevel) );
                    }

                }

			//--------------------------------------------------------------------------
			// Custom bootloader properties
			//--------------------------------------------------------------------------
            } else {
                iBlPlacementP = pobjSettings.customBlPlacementP;
                iBlSizeP = pobjSettings.customBlSizeP;
            }


			//--------------------------------------------------------------------------
			// Wait for response
			//--------------------------------------------------------------------------
            bool WaitBLReadyResponseResult = false;
            WaitBLReadyResponse( pobjPort, iTabLevel, ref WaitBLReadyResponseResult );
            if ( WaitBLReadyResponseResult == false ) {
                PortClose( pobjPort, iTabLevel );
                return;
            }


			//--------------------------------------------------------------------------
			// Setup parse settings
			//--------------------------------------------------------------------------
            clsParseSettings objParseSettings = new clsParseSettings( iBlSizeP, pobjDevice, pobjSettings.allowBlOverwrite, pobjSettings.noGoto, iBlPlacementP );

            //--------------------------------------------------------------------------
            // PIC16F
            //--------------------------------------------------------------------------
            if ( pobjDevice.family.name == "PIC16F" ) {
                clsHex16F objHex16F = (clsHex16F)pobjHex;	
                
                // Parse				
                objHex16F.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {
                    PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.0 - 0.9.1
                if ( 
                    ( strFwVer == "0.9.0" ) ||
                    ( strFwVer == "0.9.1" )
                ) {
                    bWriteSuccess = WriteFlashPIC16( pobjPort, pobjDevice, objHex16F, pobjSettings.writeProgram, pobjSettings.writeEeprom, pobjSettings.writeConfigs, iTabLevel );

                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }
            

            //--------------------------------------------------------------------------
            // PIC18F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC18F" ) {
                clsHex18F objHex18F = (clsHex18F)pobjHex;	
                
                // Parse				
                objHex18F.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {
                    PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.0 - 1.0.2
                if ( 
                    ( strFwVer == "0.9.0" ) ||
                    ( strFwVer == "0.9.2" ) ||
                    ( strFwVer == "1.0.0" ) ||
                    ( strFwVer == "1.0.1" ) ||
                    ( strFwVer == "1.0.2" )
                ) {
                    bWriteSuccess = WriteFlash8( pobjPort, pobjDevice, objHex18F, pobjSettings.writeProgram, pobjSettings.writeEeprom, pobjSettings.writeConfigs, iTabLevel );
                
                    // Write versions 1.5
                } else if ( 
                    ( strFwVer == "1.5.0" )
                ) {                    
                    bWriteSuccess = WriteFlash8( pobjPort, pobjDevice, objHex18F, pobjSettings.writeProgram, false, false, iTabLevel );
              
                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }


            //--------------------------------------------------------------------------
            // PIC18FJ
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC18FJ" ) {
                clsHex18FJ objHex18FJ = (clsHex18FJ)pobjHex;	
                
                // Parse
                objHex18FJ.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {		
			        PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.0 - 1.0.1
                if ( 
                    ( strFwVer == "0.9.0" ) ||
                    ( strFwVer == "0.9.1" ) ||
                    ( strFwVer == "1.0.0" ) ||
                    ( strFwVer == "1.0.1" )
                ) {
                    bWriteSuccess = WriteFlash8( pobjPort, pobjDevice, objHex18FJ, pobjSettings.writeProgram, false, false, iTabLevel );
            
                
                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }


            //--------------------------------------------------------------------------
            // PIC24F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24F" ) {
                clsHex24F objHex24F = (clsHex24F)pobjHex;
                
                // Parse
                objHex24F.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {
                    PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.0 - 0.9.3
                if ( 
                    ( strFwVer == "0.9.0" ) ||
                    ( strFwVer == "0.9.1" ) ||
                    ( strFwVer == "0.9.2" ) ||
                    ( strFwVer == "0.9.3" )
                ) {
                    bWriteSuccess = WriteFlash16( pobjPort, pobjDevice, objHex24F, pobjSettings.writeProgram, pobjSettings.writeEeprom, pobjSettings.writeConfigs, iTabLevel );
            
                
                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }
   

            //--------------------------------------------------------------------------
            // PIC24FJ
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24FJ" ) {
                clsHex24FJ objHex24FJ = (clsHex24FJ)pobjHex;
                
                // Parse
                objHex24FJ.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {	
				    PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.0 - 1.0.3
                if ( 
                    ( strFwVer == "0.9.0" ) ||
                    ( strFwVer == "0.9.1" ) ||
                    ( strFwVer == "1.0.0" ) ||
                    ( strFwVer == "1.0.1" ) ||
                    ( strFwVer == "1.0.2" ) ||
                    ( strFwVer == "1.0.3" ) ||
                    ( strFwVer == "1.0.4" ) 
                ) {
                    bWriteSuccess = WriteFlash16( pobjPort, pobjDevice, objHex24FJ, pobjSettings.writeProgram, false, false, iTabLevel );
            
                
                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }
                
            //--------------------------------------------------------------------------
            // PIC24H
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24H" ) {
                clsHex24H objHex24H = (clsHex24H)pobjHex;	
                
                // Parse
                objHex24H.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {		
			        PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.0 - 1.0.1
                if ( 
                    ( strFwVer == "0.9.0" ) || 
                    ( strFwVer == "0.9.1" ) ||
                    ( strFwVer == "1.0.0" ) ||
                    ( strFwVer == "1.0.1" ) ||
                    ( strFwVer == "1.0.2" ) 
                ) {
                    bWriteSuccess = WriteFlash16( pobjPort, pobjDevice, objHex24H, pobjSettings.writeProgram, false, pobjSettings.writeConfigs, iTabLevel );
                
                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }

            //--------------------------------------------------------------------------
            // dsPIC30F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "dsPIC30F" ) {
                clsHex30F objHex30F = (clsHex30F)pobjHex;
                
                // Parse
                objHex30F.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {					
                    PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write versions 0.9.4 - 1.0.1
                if ( 
					( strFwVer == "0.9.4" ) ||
					( strFwVer == "1.0.0" ) ||
                    ( strFwVer == "1.0.1" )
				) {
                    bWriteSuccess = WriteFlash30_1_0_1( pobjPort, pobjDevice, objHex30F, pobjSettings.writeProgram, pobjSettings.writeEeprom, pobjSettings.writeConfigs, iTabLevel );
                
                // Versions 2.0.0 - 2.5.0
                } else if ( 
                    ( strFwVer == "2.0.0" ) ||
                    ( strFwVer == "2.0.1" ) ||
                    ( strFwVer == "2.0.2" ) ||
                    ( strFwVer == "2.0.3" ) ||
                    ( strFwVer == "2.0.4" ) ||
                    ( strFwVer == "2.5.0" )
				) {
                    bWriteSuccess = WriteFlash16( pobjPort, pobjDevice, objHex30F, pobjSettings.writeProgram, pobjSettings.writeEeprom, pobjSettings.writeConfigs, iTabLevel );

                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }
            
            //--------------------------------------------------------------------------
            // dsPIC33F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "dsPIC33FJ" ) {
                clsHex33FJ objHex33F = (clsHex33FJ)pobjHex;	
                
                // Parse
                objHex33F.ParseHexFile( objParseSettings, iTabLevel, ref bParseResult );
                if ( bParseResult == false ) {					
                    PortClose( pobjPort, iTabLevel );
                    return;
				}

                // Write vrsions 0.9.0 - 1.0.4
                if ( 
                    ( strFwVer == "0.9.0" ) ||
                    ( strFwVer == "0.9.1" ) ||
                    ( strFwVer == "0.9.2" ) ||
                    ( strFwVer == "0.9.3" ) ||
                    ( strFwVer == "0.9.4" ) ||
                    ( strFwVer == "1.0.0" ) || 
                    ( strFwVer == "1.0.1" ) ||
                    ( strFwVer == "1.0.2" ) ||
                    ( strFwVer == "1.0.3" ) ||
                    ( strFwVer == "1.0.4" ) 
                ) {
                    bWriteSuccess = WriteFlash16( pobjPort, pobjDevice, objHex33F, pobjSettings.writeProgram, false, pobjSettings.writeConfigs, iTabLevel );
            
                
                // Unkown version
                } else {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Unknown firmware version", iTabLevel) );
                    PortClose( pobjPort, iTabLevel );
					return;
                }
            }
						
					    
			//--------------------------------------------------------------------------
			// Download completed
			//--------------------------------------------------------------------------
            pobjPort.Close();

            if ( bWriteSuccess == true ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.completed, "Download finished", iTabLevel) );	                
            }

            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progressEnded, "", 0) );
            pbResult = bWriteSuccess;
		}// Download()
		
		
		//---------------------------------------------------------------------
		// FindLoader()
		//---------------------------------------------------------------------
		static public void FindLoader( clsDevice pobjDevice, clsSerialPort pobjPort, clsDownloadSettings pobjSettings, ref Version pverFw, int iTabLevel, ref bool pbResult )
		{
            int iDeviceIdFound = -1;
            bool bGetResponseResult = false;
            bool bPortWasOpened = false;
            int iFWVerMaj = 9;
            int iFWVerMin = 9;
            int iFWVerRev = 9;

            pbResult = false;

            //-----------------------------------------------------------------
            // Test-mode
            //-----------------------------------------------------------------
            if ( testMode == true ) {
                iFWVerMaj = testFwVer.Major;
                iFWVerMin = testFwVer.Minor;
                iFWVerRev = testFwVer.Build;
                iDeviceIdFound = testDeviceID;
            
                
            //-----------------------------------------------------------------
            // Real-mode
            //-----------------------------------------------------------------
            } else {
                

			    //--------------------------------------------------------------------------
			    // Open port if needed
			    //--------------------------------------------------------------------------    
		        if ( pobjPort.isOpen == false ) {
                    bool bOpenPortResult = false;
                    PortOpen( pobjPort, pobjSettings, true, ref bOpenPortResult, iTabLevel );
                    if ( pobjPort.isOpen == false ) {
                        return;
                    }
                    bPortWasOpened = true;
                }

                
                // Empty buffers prior to hello command in case there's some crap
                pobjPort.EmptyBuffers( true, false );


			    //--------------------------------------------------------------------------
			    // Auto baudrate
			    //--------------------------------------------------------------------------                    
		        if ( pobjSettings.autoBaudrate == true ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Sending auto baudrate sync character ", iTabLevel) );
                    int iLastSend = Environment.TickCount - pobjSettings.polltime;
                    int iLastPoll = Environment.TickCount - pobjSettings.polltime;
                    int iStartTime = Environment.TickCount;                    
                    bool bOkReceived = false;
                    int iResponse;
                    bool bTimedOut = false;

                    do {  
                        // Check first byte, discard if null
                        if ( Environment.TickCount - iLastPoll >= pobjSettings.polltime/10) {     
                            iLastPoll = Environment.TickCount;                            
                            
                            if ( pobjPort.inBufferCount > 0 ) {
                                iResponse = GetResponse( pobjPort, ref bGetResponseResult );

                                if ( iResponse == 0 ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "(discarded null byte) ", -1) );
                                } else if ( iResponse != cOK ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "illegal response from auto baudrate detection(" + iResponse.ToString() + "), aborting", -1) );
                                    return;                           
                                } else {
                                    bOkReceived = true;
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                                    break;
                                }                          
                            }
                        }

                        // Send sync char                    
                       if ( Environment.TickCount - iLastSend >= pobjSettings.polltime ) {                        
                            iLastSend = Environment.TickCount;
                            pobjPort.SendByte( 0x55 );
                            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, ". ", -1) );
                            // The sync char wont be echoed
                            pobjPort.EchoClearQueue();
                        }

                        // Check for timeout
                        bTimedOut = ( Environment.TickCount - iStartTime >= pobjSettings.timeout );

                        // Handle message queue
                        Application.DoEvents();
                    } while ( 
                        bAbort == false && 
                        bTimedOut == false &&
                        bOkReceived == false
                    );
                    
                    if ( bAbort == true  ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted", -1) );
                        bAbort = false;
                        if ( bPortWasOpened == true ) pobjPort.Close();
                        return;
                    } else if ( bTimedOut == true ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "timed out", -1) );
                        if ( bPortWasOpened == true ) pobjPort.Close();
                        return;
                    }                                      
                }


			    //--------------------------------------------------------------------------
			    // Send hello to bootloader and receive device id
			    //--------------------------------------------------------------------------			    
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Searching for bl ", iTabLevel) );
                int iTimeDeviceIdReceived = 0;                
                {
                    int iLastSend = Environment.TickCount - pobjSettings.polltime;
                    int iLastPoll = Environment.TickCount - pobjSettings.polltime;
                    int iStartTime = Environment.TickCount;
                    bool bTimedOut = false;

                    do {
                        if ( Environment.TickCount - iLastPoll >= pobjSettings.polltime/10) {     
                            iLastPoll = Environment.TickCount;                            
                            
                            if ( pobjPort.inBufferCount > 0 ) {
                                iDeviceIdFound = GetResponse( pobjPort, ref bGetResponseResult );
   
                                if ( iDeviceIdFound == 0 ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "(discarded null byte) ", -1) );
                                } else {
                                    iTimeDeviceIdReceived = Environment.TickCount;
                                    break;                            
                                }                          
                            }
                        }
                        
                        // Send hello                    
                       if ( Environment.TickCount - iLastSend >= pobjSettings.polltime ) {                        
                            pobjPort.EchoClearQueue();
                            iLastSend = Environment.TickCount;
                            pobjPort.SendByte( cHello );
                            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, ". ", -1) );
                        }

                        // Check for timeout
                        bTimedOut = ( Environment.TickCount - iStartTime >= pobjSettings.timeout );

                        // Handle message queue
                        Application.DoEvents();
                    } while ( 
                        bAbort == false && 
                        bTimedOut == false
                    );
                    
                    if ( bAbort == true  ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted", -1) );
                        bAbort = false;
                        if ( bPortWasOpened == true ) pobjPort.Close();
                        return;
                    } else if ( bTimedOut == true ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "timed out", -1) );
                        if ( bPortWasOpened == true ) pobjPort.Close();
                        return;
                    }
                }

    			
			    //--------------------------------------------------------------------------
			    // Delay to receive firmware version if any
			    //--------------------------------------------------------------------------
                int iTime = 1 + 30000 / pobjPort.baudrate;
                while ( Environment.TickCount - iTimeDeviceIdReceived < iTime && pobjPort.inBufferCount < 3 ) {            
                }
    			

                //--------------------------------------------------------------------------
                // Firmware doesn´t send version, assume 0.9.4
                //--------------------------------------------------------------------------
                int iRcvVerMaj = 0, iRcvVerMinRev = 0;
                if ( pobjPort.inBufferCount == 1 ) {
                        iFWVerMaj = 0;
                        iFWVerMin = 9;
                        iFWVerRev = 4;
                
                    
                //--------------------------------------------------------------------------
                // Firmware that sends major and minor
                //--------------------------------------------------------------------------                
                } else if ( pobjPort.inBufferCount == 2 ) {
                    iRcvVerMaj = GetResponse( pobjPort, ref bGetResponseResult );
                    if ( bGetResponseResult == false ) {
                        if ( bPortWasOpened == true ) pobjPort.Close();
			            return;
                    } 
                    iFWVerMaj = ( (iRcvVerMaj & 0x70) >> 4 );
                    iFWVerMin = ( iRcvVerMaj & 0xF );
                    iFWVerRev = 0;


                //--------------------------------------------------------------------------
                // Firmware that sends major, minor and revision
                //--------------------------------------------------------------------------
                } else if ( pobjPort.inBufferCount == 3 ) {
                    iRcvVerMaj = GetResponse( pobjPort, ref bGetResponseResult );
                    if ( bGetResponseResult == false ) {
                        if ( bPortWasOpened == true ) pobjPort.Close();
			            return;
                    } 
                    iRcvVerMinRev = GetResponse( pobjPort, ref bGetResponseResult );
		            if ( bGetResponseResult == false ) {
                        if ( bPortWasOpened == true ) pobjPort.Close();
			            return;
		            }		    			
        
                    iFWVerMaj = ( iRcvVerMaj & 0x7F );
                    iFWVerMin = ( (iRcvVerMinRev & 0xF0) >> 4 );
                    iFWVerRev = ( iRcvVerMinRev & 0xF );
		        }


                //--------------------------------------------------------------------------
                // PIC18 indicated by msb in fw ver major
                //--------------------------------------------------------------------------
                if ( (iRcvVerMaj & 0x80) > 0 ) {
                    iDeviceIdFound += 300;
                }
            }//Test-/real-mode


			//--------------------------------------------------------------------------
			// Version object
			//--------------------------------------------------------------------------
            pverFw = new Version( iFWVerMaj, iFWVerMin, iFWVerRev );

            
			//--------------------------------------------------------------------------
			// Find device from device id
			//--------------------------------------------------------------------------
			clsDevice objFoundDevice = clsDevices.DeviceGet( iDeviceIdFound );

			
            //--------------------------------------------------------------------------
            // Invalid ID returned
            //--------------------------------------------------------------------------
			if ( objFoundDevice == null ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Found unknown device id(" + iDeviceIdFound.ToString() + ") fw ver. " + pverFw.ToString(), iTabLevel) );
                if ( bPortWasOpened == true ) pobjPort.Close();
				return;
			}


            //--------------------------------------------------------------------------
            // Bootloader found
            //--------------------------------------------------------------------------
            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "Found " + objFoundDevice.name + " fw ver. " + pverFw.ToString(), iTabLevel) );

            
            //--------------------------------------------------------------------------
            // Different device found from choosen
            //--------------------------------------------------------------------------
            if ( objFoundDevice.id != pobjDevice.id ) {
                OnDownloading(new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Wrong pic detected(" + objFoundDevice.id.ToString() + ")/selected(" + pobjDevice.id.ToString() + "), aborting", iTabLevel));
                if ( bPortWasOpened == true ) pobjPort.Close();
				return;
			}

			pbResult = true;

            // Close port if it was openeed
            if ( bPortWasOpened == true ) {
                pobjPort.Close();
            }
		}// FindLoader()


        //---------------------------------------------------------------------
		// WaitBLReadyResponse()
		//---------------------------------------------------------------------
		static private void WaitBLReadyResponse( clsSerialPort pobjPort, int iTabLevel, ref bool pbResult )
        {
            int iResponse = -1;
            pbResult = false;

            //--------------------------------------------------------------------------
			// Get bootloader ready response
			//--------------------------------------------------------------------------
            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Waiting for bootloader to be ready...", iTabLevel) );
			if ( testMode == false ) {
                bool bGetResponseResult = false;
                iResponse = GetResponse( pobjPort, ref bGetResponseResult);
                if ( bGetResponseResult == false ) {
				    return;
			    }
            } else {
                iResponse = cOK;
            }
		    
			if ( iResponse != cOK ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "unknown response, aborting", -1) );
				return;
			}
            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
            pbResult = true;
        }//WaitBLReadyResponse()


        //---------------------------------------------------------------------
		// GetDefaultBlPlacementP()
		//---------------------------------------------------------------------
		static private int GetDefaultBlPlacementP( clsDevice pobjDevice, Version pverFw, ref bool pbResult )
		{
            pbResult = true;
            string strVersion = pverFw.ToString();


            //--------------------------------------------------------------------------
            // PIC16F
            //--------------------------------------------------------------------------
            if ( pobjDevice.family.name == "PIC16F" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" )
                ){
                    return 6;

                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}


            //--------------------------------------------------------------------------
            // PIC18F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC18F" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" )
                ){
                    return 4;
                
                // Firmware version 0.9.2 - 1.0.2
                } else if ( 
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" )
                ){
                    return 5;
                 
                // Firmware version 1.5 - 1.5
                } else if ( 
                    ( strVersion == "1.5.0" ) ||
                    ( strVersion == "1.5.0" )
                ){
                    return 3;
    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}


            //--------------------------------------------------------------------------
            // PIC18FJ
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC18FJ" ) {
                // Firmware version 0.9.0 - 1.0.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" )
                ){
                    return 2;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	


            //--------------------------------------------------------------------------
            // PIC24F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24F" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "0.9.3" )
                ){
                    return 4;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	
   

            //--------------------------------------------------------------------------
            // PIC24FJ
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24FJ" ) {
                // Firmware version 0.9.0 - 1.0.3
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" ) ||
                    ( strVersion == "1.0.3" ) ||
                    ( strVersion == "1.0.4" ) 
                ){
                    return 2;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	
            
                
            //--------------------------------------------------------------------------
            // PIC24H
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24H" ) {
                // Firmware version 0.9.0 - 1.0.1 
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" ) 
                ){
                    return 1;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	


            //--------------------------------------------------------------------------
            // dsPIC30F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "dsPIC30F" ) {
                // Firmware version 0.9.4 - 1.0.1
                if ( 
					( strVersion == "0.9.4" ) ||
					( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" )					
				) {
                    return 3;
                
                // Firmware version 2.0.0 - 2.0.2
                } else if ( 
                    ( strVersion == "2.0.0" ) ||
                    ( strVersion == "2.0.1" ) ||
                    ( strVersion == "2.0.2" ) ||
                    ( strVersion == "2.0.3" )
                ) {
					return 4;

                // Firmware version 2.0.4 - 2.0.4
                } else if ( 
                    ( strVersion == "2.0.4" )
                ) {
					return 5;

                // Firmware version 2.5.0 - 2.5.0
                } else if ( 
                    ( strVersion == "2.5.0" )
                ) {
					return 7;

                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
                }

            
            //--------------------------------------------------------------------------
            // dsPIC33F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "dsPIC33FJ" ) {
                // Firmware version 0.9.0 - 1.0.4
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "0.9.3" ) ||
                    ( strVersion == "0.9.4" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" ) ||
                    ( strVersion == "1.0.3" ) ||
                    ( strVersion == "1.0.4" )
                ) {
                    return 1;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	
            }

            pbResult = false;
            return -1;           
        }// GetDefaultBlPlacement()


        //---------------------------------------------------------------------
		// GetDefaultBLSizeP()
		//---------------------------------------------------------------------
		static private int GetDefaultBLSizeP( clsDevice pobjDevice, Version pverFw, ref bool pbResult )
		{
            pbResult = true;
            string strVersion = pverFw.ToString();
            
            
            //--------------------------------------------------------------------------
            // PIC16F
            //--------------------------------------------------------------------------
            if ( pobjDevice.family.name == "PIC16F" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" )
                ){
                    return 6;
 
    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}


			//--------------------------------------------------------------------------
            // PIC18F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC18F" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" )
                ){
                    return 4;
                
                // Firmware version 0.9.2 - 1.0.0
                } else if ( 
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" )
                ){
                    return 5;
                 
                // Firmware version 1.5 - 1.5
                } else if ( 
                    ( strVersion == "1.5.0" ) ||
                    ( strVersion == "1.5.0" )
                ){
                    return 3;
    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}


            //--------------------------------------------------------------------------
            // PIC18FJ
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC18FJ" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" )
                ){
                    return 1;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	


            //--------------------------------------------------------------------------
            // PIC24F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24F" ) {
                // Firmware version 0.9.0 - 0.9.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "0.9.3" )
                ){
                    return 4;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	
   

            //--------------------------------------------------------------------------
            // PIC24FJ
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24FJ" ) {
                // Firmware version 0.9.0 - 1.0.3
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" ) ||
                    ( strVersion == "1.0.3" ) ||
                    ( strVersion == "1.0.4" ) 
                ){
                    return 1;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	
            
                
            //--------------------------------------------------------------------------
            // PIC24H
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "PIC24H" ) {
                // Firmware version 0.9.0 - 1.0.1 
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" ) 
                ){
                    return 1;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	


            //--------------------------------------------------------------------------
            // dsPIC30F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "dsPIC30F" ) {
                // Firmware version 0.9.4 - 1.0.1
                if ( 
					( strVersion == "0.9.4" ) ||
					( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" )					
				) {
                    return 3;
                
                // Firmware version 2.0.0 - 2.0.2
                } else if ( 
                    ( strVersion == "2.0.0" ) ||
                    ( strVersion == "2.0.1" ) ||
                    ( strVersion == "2.0.2" ) ||
                    ( strVersion == "2.0.3" )
                ) {
					return 4;
                
                // Firmware version 2.0.4 - 2.0.4
                } else if ( 
                    ( strVersion == "2.0.4" )
                ) {
					return 5;

                // Firmware version 2.5.0 - 2.5.0
                } else if ( 
                    ( strVersion == "2.5.0" )
                ) {
					return 7;

                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
                }

            
            //--------------------------------------------------------------------------
            // dsPIC33F
            //--------------------------------------------------------------------------
            } else if ( pobjDevice.family.name == "dsPIC33FJ" ) {
                // Firmware version 0.9.0 - 1.0.1
                if ( 
                    ( strVersion == "0.9.0" ) ||
                    ( strVersion == "0.9.1" ) ||
                    ( strVersion == "0.9.2" ) ||
                    ( strVersion == "0.9.3" ) ||
                    ( strVersion == "0.9.4" ) ||
                    ( strVersion == "1.0.0" ) ||
                    ( strVersion == "1.0.1" ) ||
                    ( strVersion == "1.0.2" ) ||
                    ( strVersion == "1.0.3" ) ||
                    ( strVersion == "1.0.4" )
                ) {
                    return 1;
                    
                // Unknown firmware version
                } else {
					pbResult = false;
                    return 0;
				}	
            }

            pbResult = false;
            return -1;
		}// GetDefaultBLSizeP()
		
		
		//---------------------------------------------------------------------
		// GetResponse()
		//---------------------------------------------------------------------
		static public int GetResponse( clsSerialPort pobjPort, ref bool pbResult )
		{
            int iResponse = -1;
            pbResult = false;

            // Test-mode
            if ( testMode == true ) {
                pbResult = true;
                return testResponse;
            }

            //
			int iStartTime = System.Environment.TickCount;
		    while ( pobjPort.inBufferCount < 1 && System.Environment.TickCount - iStartTime < 2000 && bAbort == false ) {
				Application.DoEvents();
			}

            // Abort
            if ( bAbort == true ) {
                return -1;
            }
			
			// Timeout
			if ( pobjPort.inBufferCount < 1 ) {     
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "response timed out", -1) );
				return -1;
			}			
		    
			pobjPort.ReadByte( ref iResponse );
			pbResult = true;
            return iResponse;
		}// GetResponse()

        
        //---------------------------------------------------------------------
		// PortClose()
		//---------------------------------------------------------------------
		static private void PortClose( clsSerialPort pobjPort, int piTabLevel )
		{
            if ( debugMode == true ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Closing port", 0) );  
            }
            pobjPort.dtrEnable = false;
            pobjPort.rtsEnable = false;
            pobjPort.Close();
		}// PortClose()


		//---------------------------------------------------------------------
		// PortOpen()
		//---------------------------------------------------------------------
		static private void PortOpen( clsSerialPort pobjPort, clsDownloadSettings pobjSettings, bool pbDoActivation, ref bool pbResult, int piTabLevel )
		{
            pbResult = false;
            if ( debugMode == true ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Opening port", piTabLevel) );  
            }

            if ( pbDoActivation == true ) {
                // Dtr activation
                if ( pobjSettings.activateDtr == true ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Bringing dtr high...", piTabLevel) );  
                    try {
			            pobjPort.objPort.DtrEnable = true;
                    } catch {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "failed", -1) );  
                        return;
                    }
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );  
                } else {
                    pobjPort.objPort.DtrEnable = false;
                }

                // Rts activation
                if ( pobjSettings.activateRts == true ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Bringing rts high...", piTabLevel) );  
                    try {
			            pobjPort.objPort.RtsEnable = true;
                    } catch {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "failed", -1) );  
                        return;
                    }
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );  
                } else {
                    pobjPort.objPort.RtsEnable = false;
                }
            }

            // Allready open?
            if ( pobjPort.isOpen == true ) {
                pbResult = true;
                return;
            }
            
            try {
                pobjPort.Open();
            } catch {
            }

            if ( pobjPort.isOpen == false ) {
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "Failed to open port " + pobjPort.portName, piTabLevel) );  
            }

            pbResult = pobjPort.isOpen;
		}// PortOpen()


		//---------------------------------------------------------------------
		// ProcessWriteResponse()
		//---------------------------------------------------------------------
		static private void ProcessWriteResponse( clsSerialPort pobjPort, ref int piRetries, int piTabLevel, string pstrWriteType, ref bool pbRetry, ref bool pbResult )
		{
            int iResponse = -1;
            pbResult = false;

            // Get response
            bool bGetResponseResult = false;
            iResponse = GetResponse( pobjPort, ref bGetResponseResult );
            if ( bGetResponseResult == false ) {
                return;
            }

            // Check response
		    if ( iResponse != cOK ) {
			    if ( ++piRetries == 3 ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "download failed", -1) );
				    return;
			    } else {
				    if ( iResponse == cChecksumErr ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "checksum error," + piRetries.ToString() + " try", -1) );
                        pbRetry = true;
                    } else if ( iResponse == cVerifyErr ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "verification error," + piRetries.ToString() + " try", -1) );
                        pbRetry = true;
                    } else if ( iResponse == cBlProtTrip ) {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "bl protection tripped", -1) );
                        pbRetry = false;
                        return;
                    } else {
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "unknown response 0x" + iResponse.ToString("X2") + ", " + piRetries.ToString() + " try", -1) );
                        pbRetry = true;
                    }
                    
                    if ( debugMode == false ) {
				        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing " + pstrWriteType + "...", piTabLevel) );
                    }
			    }
		    } else {
                pbRetry = false;
            }

            pbResult = true;
		}// ProcessWriteResponse()


		//---------------------------------------------------------------------
		// ResetDevice()
		//---------------------------------------------------------------------
		static private void ResetDevice( clsSerialPort pobjPort, clsDownloadSettings pobjSettings, int iTabLevel, ref bool pbResult )
		{
            pbResult = false;

            //--------------------------------------------------------------------------
		    // Reset - command
		    //--------------------------------------------------------------------------
            if ( pobjSettings.resetCommand == true ) {
                // Parse resetcommand
                bool bSplitResult = false;
                byte[] bBytes = pobjPort.SplitHexStringToBytes( pobjSettings.resetCommandSequence, ref bSplitResult );
                if ( bSplitResult == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "couldn't parse reset command sequence", -1) );
                    return;
                }

                // Open port
                int iOldBaudrate = pobjPort.baudrate;                
                pobjPort.baudrate = pobjSettings.resetBaudrate;                
                bool bOpenPortResult = false;
                PortOpen( pobjPort, pobjSettings, false, ref bOpenPortResult, iTabLevel );  
                if ( bOpenPortResult == false ) {
                    return;
                }
                
                // Send reset command and close port         
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Sending reset command...", iTabLevel) );  
                pobjPort.SendBytes( ref bBytes, bBytes.Length );
                while ( pobjPort.outBufferCount > 0 );
                pobjPort.Close();
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );

                // Restore baudrate
                pobjPort.baudrate = iOldBaudrate;

                SleepResponsive( pobjSettings.resetTime );


		    //--------------------------------------------------------------------------
		    // Reset - dtr
		    //--------------------------------------------------------------------------            
            } else if ( pobjSettings.resetDtr == true ) {
                // Notify user
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Resetting by dtr...", iTabLevel) );
                // Open port
                bool bPortOpenResult = false;
                PortOpen( pobjPort, pobjSettings, false, ref bPortOpenResult, iTabLevel );
                if ( bPortOpenResult == false ) {
                    return;
                }
                // Do reset
                pobjPort.dtrEnable = true;
                SleepResponsive( pobjSettings.resetTime );
                pobjPort.dtrEnable = false;
                // Close port
                PortClose( pobjPort, iTabLevel );
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );


		    //--------------------------------------------------------------------------
		    // Reset - rts
		    //--------------------------------------------------------------------------           
            } else if ( pobjSettings.resetRts == true ) {
                // Notify user
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Resetting by rts...", iTabLevel) );
                // Open port
                bool bPortOpenResult = false;
                PortOpen( pobjPort, pobjSettings, false, ref bPortOpenResult, iTabLevel );
                if ( bPortOpenResult == false ) {
                    return;
                }                
                // Do reset
                pobjPort.rtsEnable = true;
                SleepResponsive( pobjSettings.resetTime );
                pobjPort.rtsEnable = false;
                // Close port
                PortClose( pobjPort, iTabLevel );
                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
            }

            pbResult = true;
        }//ResetDevice()


        //---------------------------------------------------------------------
		// SleepResponsive()
		//---------------------------------------------------------------------
		static private void SleepResponsive( int piTimeMs )
		{
            int iStartTime = Environment.TickCount;
            while ( Environment.TickCount - iStartTime < piTimeMs ) {
                Application.DoEvents();
            }
        }//SleepResponsive()


        #region WriteFlash
        
        
        //---------------------------------------------------------------------
        // WriteFlashPIC16()
        //---------------------------------------------------------------------
        static private bool WriteFlashPIC16(clsSerialPort pobjPort, clsDevice pobjDevice, clsHex pobjHex, bool pbWriteProgram, bool pbWriteEeprom, bool pbWriteConfigs, int iTabLevel)
        {
            //--------------------------------------------------------------------------
            // Variables
            //--------------------------------------------------------------------------
            int iChecksum;
            int iSum;
            int iAddrP;
            int iRetries;
            byte[] bBuffer = new byte[256];
            int iBufferIndx;
            byte bAddrU, bAddrH, bAddrL, bSize, bProg;
            bool bRetry = false;
            int iByteToWrite = pobjHex.BytesToWrite( pobjDevice, pbWriteProgram, pbWriteEeprom, pbWriteConfigs );
            int iBytesWritten = 0;
            bool bProcessWriteResponseResult = false;


            //--------------------------------------------------------------------------
            // Write flash
            //--------------------------------------------------------------------------
            int iRow;

            if ( pbWriteProgram == true ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing flash...", iTabLevel) );
                }                

                //-------------------------------------------------------------
                // Iterate pages
                //-------------------------------------------------------------
                for ( int iPage = 0; iPage < pobjHex.bProgPageUsed.Length; iPage++ ) {
                    if ( pobjHex.bProgPageUsed[iPage] == true ) {
                        iRetries = 0;
                        do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Erasing page " + iPage.ToString() + "...", iTabLevel) );
                            }

                            //---------------------------------------------------------
                            // Erase page
                            //---------------------------------------------------------
                            iSum = 0;
                            iBufferIndx = 0;

                            // Calculate
                            iAddrP = iPage * pobjDevice.pageSizeR * pobjDevice.rowsizeW;//24-bit address in memory to write to
                            bAddrU = Convert.ToByte((iAddrP & 0xff0000) >> 16);				//address upper byte			               
                            bAddrH = Convert.ToByte((iAddrP & 0x00ff00) >> 8);				//address high byte
                            bAddrL = Convert.ToByte((iAddrP & 0x0000ff) >> 0);				//address low byte
                            bSize = Convert.ToByte(1/*checksum*/ );

                            //
                            bBuffer[iBufferIndx++] = bAddrU; iSum += bAddrU;
                            bBuffer[iBufferIndx++] = bAddrH; iSum += bAddrH;
                            bBuffer[iBufferIndx++] = bAddrL; iSum += bAddrL;
                            bBuffer[iBufferIndx++] = cCmdErasePage; iSum += cCmdErasePage;
                            bBuffer[iBufferIndx++] = bSize; iSum += bSize;

                            // Calculate checksum and add to buffer
                            iSum %= 256;
                            iChecksum = Convert.ToInt16((256 - iSum) % 256);
                            bBuffer[iBufferIndx++] = Convert.ToByte(iChecksum);

                            // Send address, command, packetsize & checksum
                            pobjPort.SendBytes(ref bBuffer, iBufferIndx);

                            // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }

                            //
                            Application.DoEvents();
                        } while ( bRetry == true && bAbort == false);


                        //---------------------------------------------------------
                        // Iterate rows, write all rows in page even if not in hex-file
                        //---------------------------------------------------------
                        for ( iRow = iPage * pobjDevice.pageSizeR; iRow < iPage * pobjDevice.pageSizeR + pobjDevice.pageSizeR; iRow++ ) {
                            iRetries = 0;
                            do {
                                if ( clsds30Loader.debugMode ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing row " + iRow.ToString() + "...", iTabLevel) );
                                }

                                iBufferIndx = 0;
                                iSum = 0;

                                // Calculate
                                iAddrP = iRow * pobjDevice.rowsizeW;				//24-bit address in memory to write to
                                bAddrU = Convert.ToByte((iAddrP & 0xff0000) >> 16);	    //address upper byte			               
                                bAddrH = Convert.ToByte((iAddrP & 0x00ff00) >> 8);	    //address high byte
                                bAddrL = Convert.ToByte((iAddrP & 0x0000ff));			//address low byte
                                bSize = Convert.ToByte(pobjDevice.rowsizeW * 2 + 1/*checksum*/ );

                                //
                                bBuffer[iBufferIndx++] = bAddrU; iSum += bAddrU;
                                bBuffer[iBufferIndx++] = bAddrH; iSum += bAddrH;
                                bBuffer[iBufferIndx++] = bAddrL; iSum += bAddrL;
                                bBuffer[iBufferIndx++] = cCmdWriteRow; iSum += cCmdWriteRow;
                                bBuffer[iBufferIndx++] = bSize; iSum += bSize;

						        // Buffer row
                                for ( int iWord = 0; iWord < pobjDevice.rowsizeW; iWord++ ) {
                                    // Low byte
								    bProg = Convert.ToByte( pobjHex.iProgMem[iAddrP*2 + iWord*2] );
							        bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
                                    // High byte
                                    bProg = Convert.ToByte( 0x3f & pobjHex.iProgMem[iAddrP*2 + iWord*2+1] );    //mask, word is 14 bits only
							        bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
						        }

                                // Calculate checksum and add to buffer
                                iSum %= 256;
                                iChecksum = Convert.ToInt16((256 - iSum) % 256);
                                bBuffer[iBufferIndx++] = Convert.ToByte(iChecksum);

                                // Send address, command, packetsize, row & checksum
                                pobjPort.SendBytes( ref bBuffer, iBufferIndx );

                                // Get response, checksum ok
                                ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                                if ( bProcessWriteResponseResult == false ) {
                                    return false;
                                } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }

                                //
                                Application.DoEvents();
                            } while ( bRetry == true && bAbort == false );

                            //
                            if (bAbort == true) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
                                return false;
                            }

                            //
                            iBytesWritten += pobjDevice.rowsizeW * 2;
                            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );

                        }//for ( iRow = 0; iRow < iRowUsedBufferSize; iRow++ ) {

                    } //if ( bPageUsed[iIter] == true ) {

                }//for ( iPage = 0; iPage < iPageUsedBufferSize; iPage++ ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteProgram == true ) {


            //--------------------------------------------------------------------------
            // Write Eeprom
            //--------------------------------------------------------------------------
            if ( pbWriteEeprom == true ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom...", iTabLevel) );
                }
				
			    for ( int iEepromWord = 0; iEepromWord < pobjDevice.EepromSizeB; iEepromWord++ ) {					
				    if ( pobjHex.bEEWordUsed[iEepromWord] == true ) {
						iRetries = 0;
					    do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom byte " + iEepromWord.ToString() + " (0x" + pobjHex.iEEMem[iEepromWord].ToString("X2") + ")...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;
    		                
		                    // Calculate
		                    iAddrP = iEepromWord;		                            //24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( 1 + 1/*checksum*/ );
    						
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
                            bBuffer[ iBufferIndx++ ] = cCmdWriteEEWord; iSum += cCmdWriteEEWord;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;

							// Eeprom byte
						    bProg = Convert.ToByte( pobjHex.iEEMem[iEepromWord] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
    		                
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
    						
						    // Send address, command, size, Eeprom byte & checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
    		                
		                    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "Eeprom", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }
    		                
		                    //
						    Application.DoEvents();
					    } while ( bRetry == true && bAbort == false );
    		            
		                //
					    if ( bAbort == true ) {
						    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
						    return false;
					    }
    		            
    		            //
					    iBytesWritten += 1;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
				     }//if (  bEEWordUsed[iEepromWord] == true ) {				     
				     
			    }//for ( iIter = 0; iIter < pobjDevice.EepromSizeB/2; iIter++ ) {

			    if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteEeprom == true ) {

            
            //--------------------------------------------------------------------------
			// Write Config Bits
			//--------------------------------------------------------------------------
            if ( pbWriteConfigs == true ) {
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing configs...", iTabLevel) );
                }

				for ( int iConfig = 0; iConfig < pobjHex.bConfigWordUsed.Length; iConfig++ ) {
					if ( pobjHex.bConfigWordUsed[iConfig] == true ) {
                        iRetries = 0;
						do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing config byte " + iConfig.ToString() + " (0x" + pobjHex.iConfigMem[iConfig].ToString("X2") + ")...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;						
						
						    // Calculate
						    iAddrP = 0x300000 + iConfig;               				//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( 1 + 1/*checksum*/ );
							
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
						    bBuffer[ iBufferIndx++ ] = cCmdWriteConfig;iSum += cCmdWriteConfig;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;	
		                    
						    // Buffer config word
						    bProg = Convert.ToByte( pobjHex.iConfigMem[ iConfig ] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;	
							
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
							
						    // Send row+checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
		                    
						    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "config", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }

                            Application.DoEvents();
						} while ( bRetry == true && bAbort == false );
		                
						//
						if ( bAbort == true ) {
							OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
							return false;
						}
						
                        //
						iBytesWritten += 1;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
					}//if ( bConfigUsed[iIter] == true ) {
					
				}//for ( iIter = 0;  iIter < iConfigUsedBufferSize; iIter++ ) { 
				
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                }
			}//if ( bWriteConfigs == true )


            //--------------------------------------------------------------------------
            // Write completed
            //--------------------------------------------------------------------------
            return true;
        }// WriteFlashPIC16()


        //---------------------------------------------------------------------
        // WriteFlash8()
        //---------------------------------------------------------------------
        static private bool WriteFlash8(clsSerialPort pobjPort, clsDevice pobjDevice, clsHex pobjHex, bool pbWriteProgram, bool pbWriteEeprom, bool pbWriteConfigs, int iTabLevel)
        {
            //--------------------------------------------------------------------------
            // Variables
            //--------------------------------------------------------------------------
            int iChecksum;
            int iSum;
            int iAddrP;
            int iRetries;
            byte[] bBuffer = new byte[256];
            int iBufferIndx;
            byte bAddrU, bAddrH, bAddrL, bSize, bProg;
            bool bRetry = false;
            int iByteToWrite = pobjHex.BytesToWrite( pobjDevice, pbWriteProgram, pbWriteEeprom, pbWriteConfigs );
            int iBytesWritten = 0;
            bool bProcessWriteResponseResult = false;


            //--------------------------------------------------------------------------
            // Write flash
            //--------------------------------------------------------------------------
            int iRow;

            if ( pbWriteProgram == true ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing flash...", iTabLevel) );
                }                

                //-------------------------------------------------------------
                // Iterate pages
                //-------------------------------------------------------------
                for ( int iPage = 0; iPage < pobjHex.bProgPageUsed.Length; iPage++ ) {
                    if ( pobjHex.bProgPageUsed[iPage] == true ) {
                        iRetries = 0;
                        do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Erasing page " + iPage.ToString() + "...", iTabLevel) );
                            }

                            //---------------------------------------------------------
                            // Erase page
                            //---------------------------------------------------------
                            iSum = 0;
                            iBufferIndx = 0;

                            // Calculate
                            iAddrP = iPage * pobjDevice.pageSizeR * pobjDevice.rowsizeW * 2;//24-bit address in memory to write to
                            bAddrU = Convert.ToByte((iAddrP & 0xff0000) >> 16);				//address upper byte			               
                            bAddrH = Convert.ToByte((iAddrP & 0x00ff00) >> 8);				//address high byte
                            bAddrL = Convert.ToByte((iAddrP & 0x0000ff) >> 0);				//address low byte
                            bSize = Convert.ToByte(1/*checksum*/ );

                            //
                            bBuffer[iBufferIndx++] = bAddrU; iSum += bAddrU;
                            bBuffer[iBufferIndx++] = bAddrH; iSum += bAddrH;
                            bBuffer[iBufferIndx++] = bAddrL; iSum += bAddrL;
                            bBuffer[iBufferIndx++] = cCmdErasePage; iSum += cCmdErasePage;
                            bBuffer[iBufferIndx++] = bSize; iSum += bSize;

                            // Calculate checksum and add to buffer
                            iSum %= 256;
                            iChecksum = Convert.ToInt16((256 - iSum) % 256);
                            bBuffer[iBufferIndx++] = Convert.ToByte(iChecksum);

                            // Send address, command, packetsize & checksum
                            pobjPort.SendBytes(ref bBuffer, iBufferIndx);

                            // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }

                            //
                            Application.DoEvents();
                        } while ( bRetry == true && bAbort == false);


                        //---------------------------------------------------------
                        // Iterate rows, write all rows in page even if not in hex-file
                        //---------------------------------------------------------
                        for ( iRow = iPage * pobjDevice.pageSizeR; iRow < iPage * pobjDevice.pageSizeR + pobjDevice.pageSizeR; iRow++ ) {
                            iRetries = 0;
                            do {
                                if ( clsds30Loader.debugMode ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing row " + iRow.ToString() + "...", iTabLevel) );
                                }

                                iBufferIndx = 0;
                                iSum = 0;

                                // Calculate
                                iAddrP = 2 * iRow * pobjDevice.rowsizeW;				//24-bit address in memory to write to
                                bAddrU = Convert.ToByte((iAddrP & 0xff0000) >> 16);	    //address upper byte			               
                                bAddrH = Convert.ToByte((iAddrP & 0x00ff00) >> 8);	    //address high byte
                                bAddrL = Convert.ToByte((iAddrP & 0x0000ff));			//address low byte
                                bSize = Convert.ToByte(pobjDevice.rowsizeW * 2 + 1/*checksum*/ );

                                //
                                bBuffer[iBufferIndx++] = bAddrU; iSum += bAddrU;
                                bBuffer[iBufferIndx++] = bAddrH; iSum += bAddrH;
                                bBuffer[iBufferIndx++] = bAddrL; iSum += bAddrL;
                                bBuffer[iBufferIndx++] = cCmdWriteRow; iSum += cCmdWriteRow;
                                bBuffer[iBufferIndx++] = bSize; iSum += bSize;

						        // Buffer row
						        for ( int iByte = 0; iByte < pobjDevice.rowsizeW * 2; iByte++ ) {
								    bProg = Convert.ToByte( pobjHex.iProgMem[iAddrP + iByte] );
							        bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
						        }

                                // Calculate checksum and add to buffer
                                iSum %= 256;
                                iChecksum = Convert.ToInt16((256 - iSum) % 256);
                                bBuffer[iBufferIndx++] = Convert.ToByte(iChecksum);

                                // Send address, command, packetsize, row & checksum
                                pobjPort.SendBytes( ref bBuffer, iBufferIndx );

                                // Get response, checksum ok
                                ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                                if ( bProcessWriteResponseResult == false ) {
                                    return false;
                                } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }

                                //
                                Application.DoEvents();
                            } while ( bRetry == true && bAbort == false );

                            //
                            if (bAbort == true) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
                                return false;
                            }

                            //
                            iBytesWritten += pobjDevice.rowsizeW * 2;
                            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );

                        }//for ( iRow = 0; iRow < iRowUsedBufferSize; iRow++ ) {

                    } //if ( bPageUsed[iIter] == true ) {

                }//for ( iPage = 0; iPage < iPageUsedBufferSize; iPage++ ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteProgram == true ) {


            //--------------------------------------------------------------------------
            // Write Eeprom
            //--------------------------------------------------------------------------
            if ( pbWriteEeprom == true ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom...", iTabLevel) );
                }
				
			    for ( int iEepromWord = 0; iEepromWord < pobjDevice.EepromSizeB; iEepromWord++ ) {					
				    if ( pobjHex.bEEWordUsed[iEepromWord] == true ) {
						iRetries = 0;
					    do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom byte " + iEepromWord.ToString() + " (0x" + pobjHex.iEEMem[iEepromWord].ToString("X2") + ")...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;
    		                
		                    // Calculate
		                    iAddrP = iEepromWord;		                            //24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( 1 + 1/*checksum*/ );
    						
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
                            bBuffer[ iBufferIndx++ ] = cCmdWriteEEWord; iSum += cCmdWriteEEWord;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;

							// Eeprom byte
						    bProg = Convert.ToByte( pobjHex.iEEMem[iEepromWord] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
    		                
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
    						
						    // Send address, command, size, Eeprom byte & checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
    		                
		                    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "Eeprom", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }
    		                
		                    //
						    Application.DoEvents();
					    } while ( bRetry == true && bAbort == false );
    		            
		                //
					    if ( bAbort == true ) {
						    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
						    return false;
					    }
    		            
    		            //
					    iBytesWritten += 1;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
				     }//if (  bEEWordUsed[iEepromWord] == true ) {				     
				     
			    }//for ( iIter = 0; iIter < pobjDevice.EepromSizeB/2; iIter++ ) {

			    if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteEeprom == true ) {

            
            //--------------------------------------------------------------------------
			// Write Config Bits
			//--------------------------------------------------------------------------
            if ( pbWriteConfigs == true ) {
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing configs...", iTabLevel) );
                }

				for ( int iConfig = 0; iConfig < pobjHex.bConfigWordUsed.Length; iConfig++ ) {
					if ( pobjHex.bConfigWordUsed[iConfig] == true ) {
                        iRetries = 0;
						do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing config byte " + iConfig.ToString() + " (0x" + pobjHex.iConfigMem[iConfig].ToString("X2") + ")...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;						
						
						    // Calculate
						    iAddrP = 0x300000 + iConfig;               				//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( 1 + 1/*checksum*/ );
							
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
						    bBuffer[ iBufferIndx++ ] = cCmdWriteConfig;iSum += cCmdWriteConfig;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;	
		                    
						    // Buffer config word
						    bProg = Convert.ToByte( pobjHex.iConfigMem[ iConfig ] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;	
							
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
							
						    // Send row+checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
		                    
						    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "config", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                            }

                            Application.DoEvents();
						} while ( bRetry == true && bAbort == false );
		                
						//
						if ( bAbort == true ) {
							OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
							return false;
						}
						
                        //
						iBytesWritten += 1;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
					}//if ( bConfigUsed[iIter] == true ) {
					
				}//for ( iIter = 0;  iIter < iConfigUsedBufferSize; iIter++ ) { 
				
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                }
			}//if ( bWriteConfigs == true )


            //--------------------------------------------------------------------------
            // Write completed
            //--------------------------------------------------------------------------
            return true;
        }// WriteFlash8()

        
        //---------------------------------------------------------------------
		// WriteFlash30_1_0_1()
		//---------------------------------------------------------------------
		static private bool WriteFlash30_1_0_1( clsSerialPort pobjPort, clsDevice pobjDevice, clsHex30F pobjHex, bool pbWriteProgram, bool pbWriteEeprom, bool pbWriteConfigs, int iTabLevel )
		{
			//--------------------------------------------------------------------------
			// Variables
			//--------------------------------------------------------------------------
			int iIter2;
			int iChecksum;
			int iSum;
			int iAddrP;
			int iRetries;
			byte [] bBuffer = new byte [256];
			int iBufferIndx;
			byte bAddrU, bAddrH, bAddrL, bSize, bProg;			
            bool bRetry = false;
            int iByteToWrite = pobjHex.BytesToWrite( pobjDevice, pbWriteProgram, pbWriteEeprom, pbWriteConfigs );
            int iBytesWritten = 0;
			bool bProcessWriteResponseResult = false;
		

			//--------------------------------------------------------------------------
			// Write flash
			//--------------------------------------------------------------------------
			if ( pbWriteProgram == true ) {
                if ( debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing flash...", iTabLevel) );
                }
				
				
				//-------------------------------------------------------------
				// Iterate rows
				//-------------------------------------------------------------
				for ( int iRow = 0; iRow < clsHex30F.iProgRowUsedBufferSize; iRow++ ) {
                    if ( pobjHex.bProgRowUsed[iRow] == true ) {
                        iRetries = 0;
					    do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing row " + iRow.ToString() + "...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;                                
			                
						    // Calculate
						    iAddrP = iRow * (pobjDevice.rowsizeW * 2);				//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( pobjDevice.rowsizeW * 3 + 1/*checksum*/ );
							
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;
							
						    // Calculate startindex in hex-buffer
						    iAddrP = iRow * pobjDevice.rowsizeW * 3;	
											
						    // Buffer row, upper-low-high, ordered in ParseHex()
						    for ( int iByte = 0; iByte < pobjDevice.rowsizeW*3; iByte++ ) {
								bProg = Convert.ToByte( pobjHex.iProgMem[iAddrP + iByte] );
							    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
						    }				    
		                
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
							
						    // Send address, command, packetsize, row & checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
			                
						    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                            }
			                
						    //
						    Application.DoEvents();
					    } while ( bRetry == true && bAbort == false );
				            
					    //
					    if ( bAbort == true ) {
						    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
						    return false;
					    }
			            
					    //
					    iBytesWritten += pobjDevice.rowsizeW * 3;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
				
				    } //if ( bProgRowUsed[iIter] == true ) {
                }//for ( iRow = 0; iRow < iProgRowUsedBufferSize; iRow++ ) {
									
                if ( debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
			}//if ( pbWriteProgram == true ) {


		    //--------------------------------------------------------------------------
            // Write Eeprom
            //--------------------------------------------------------------------------
            if ( pbWriteEeprom == true ) {
                if ( debugMode == false ) {
			        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom...", iTabLevel) );
                }
				
			    for ( int iEepromWord = 0; iEepromWord < pobjDevice.EepromSizeB/2; iEepromWord++ ) {					
				    if (  pobjHex.bEEWordUsed[iEepromWord] == true ) {
						iRetries = 0;
						
					    do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom word " + iEepromWord.ToString() + " (0x" + ((pobjHex.iEEMem[iEepromWord*2+1]<<8)+pobjHex.iEEMem[iEepromWord * 2]).ToString("X4") + ")...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;
    		                
		                    // Calculate
		                    iAddrP = pobjDevice.EepromStartAddress + iEepromWord * 2;		//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( 2 + 1/*checksum*/ );
    						
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;

							// Low byte
						    bProg = Convert.ToByte( pobjHex.iEEMem[iEepromWord*2 + 0] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
							
							// High byte								    
						    bProg = Convert.ToByte( pobjHex.iEEMem[iEepromWord*2 + 1] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
    		                
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
    						
						    // Send row+checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
    		                
		                    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "Eeprom", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                            }
    		                
		                    //
						    Application.DoEvents();
					    } while ( bRetry == true && bAbort == false );
    		            
		                //
					    if ( bAbort == true ) {
						    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
						    return false;
					    }
    		            
    		            //
					    iBytesWritten += 2;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
				     }//if (  bEEWordUsed[iEepromWord] == true ) {				     
				     
			    }//for ( iIter = 0; iIter < pobjDevice.EepromSizeB/2; iIter++ ) {
			    
                if ( debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteEeprom == true ) {


			//--------------------------------------------------------------------------
			// Write Config Bits
			//--------------------------------------------------------------------------
			if ( pbWriteConfigs == true ) {
                if ( debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing configs...", iTabLevel) );
                }
								
				for ( int iConfig = 0;  iConfig < clsHex30F.iConfigWordsUsedBufferSize; iConfig++ ) {
					if ( pobjHex.bConfigWordUsed[iConfig] == true ) {
						iRetries = 0;
						
						do {
                            if ( clsds30Loader.debugMode == true ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing config word " + iConfig.ToString() + " (0x" + ((pobjHex.iConfigMem[iConfig * 2+1]<<8)+pobjHex.iConfigMem[iConfig * 2]).ToString("X4") + ")...", iTabLevel) );
                            }

							iBufferIndx = 0;
							iSum = 0;						
						
							// Calculate
							iAddrP = 0xF80000 + iConfig * 2;						//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
							bSize = Convert.ToByte( 2 + 1/*checksum*/ );
							
							//
							bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
							bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
							bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
							bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;	
		                    
							// Buffer config word
							for ( iIter2 = 0; iIter2 < 2; iIter2++ ) {
								bProg = Convert.ToByte( pobjHex.iConfigMem[ iConfig * 2 + iIter2 ] );
								bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;	
							}
							
							// Calculate checksum and add to buffer
							iSum %= 256;
							iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
							bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
							
							// Send row+checksum
							pobjPort.SendBytes( ref bBuffer, iBufferIndx );
		                    
							// Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "config", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                            }

                            //
                            Application.DoEvents();
						} while ( bRetry == true && bAbort == false );
		                
						//
						if ( bAbort == true ) {
							OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
							return false;
						}
						
                        //
						iBytesWritten += 2;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
					}//if ( bConfigWordUsed[iIter] == true ) {
					
				}//for ( iIter = 0;  iIter < iConfigUsedBufferSize; iIter++ ) { 
				
                if ( debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                }
			}//if ( bWriteConfigs == true )

		    
			//--------------------------------------------------------------------------
			// Write completed
			//--------------------------------------------------------------------------
			return true;
		}// WriteFlash30_1_0_1()


        //---------------------------------------------------------------------
        // WriteFlash16()
        //---------------------------------------------------------------------
        static private bool WriteFlash16(clsSerialPort pobjPort, clsDevice pobjDevice, clsHex pobjHex, bool pbWriteProgram, bool pbWriteEeprom, bool pbWriteConfigs, int iTabLevel)
        {
            //--------------------------------------------------------------------------
            // Variables
            //--------------------------------------------------------------------------
            int iChecksum;
            int iSum;
            int iAddrP;
            int iRetries;
            byte[] bBuffer = new byte[256];
            int iBufferIndx;
            byte bAddrU, bAddrH, bAddrL, bSize, bProg;
            bool bRetry = false;
            int iByteToWrite = pobjHex.BytesToWrite( pobjDevice, pbWriteProgram, pbWriteEeprom, pbWriteConfigs );
            int iBytesWritten = 0;
            bool bProcessWriteResponseResult = false;


            //--------------------------------------------------------------------------
            // Write flash
            //--------------------------------------------------------------------------
            int iRow;

            if ( pbWriteProgram == true ) {
                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing flash...", iTabLevel) );
                }                

                //-------------------------------------------------------------
                // Iterate pages
                //-------------------------------------------------------------
                for ( int iPage = 0; iPage < pobjHex.bProgPageUsed.Length; iPage++ ) {
                    if ( pobjHex.bProgPageUsed[iPage] == true ) {
                        if ( pobjDevice.pageSizeR > 1 ) {
                            iRetries = 0;
                            do {
                                if ( clsds30Loader.debugMode ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Erasing page " + iPage.ToString() + "...", iTabLevel) );
                                }

                                //---------------------------------------------------------
                                // Erase page
                                //---------------------------------------------------------
                                iSum = 0;
                                iBufferIndx = 0;

                                // Calculate
                                iAddrP = iPage * pobjDevice.pageSizeR * pobjDevice.rowsizeW * 2;	//24-bit address in memory to write to
                                bAddrU = Convert.ToByte((iAddrP & 0xff0000) >> 16);				//address upper byte			               
                                bAddrH = Convert.ToByte((iAddrP & 0x00ff00) >> 8);				//address high byte
                                bAddrL = Convert.ToByte((iAddrP & 0x0000ff) >> 0);				//address low byte
                                bSize = Convert.ToByte(1/*checksum*/ );

                                //
                                bBuffer[iBufferIndx++] = bAddrU; iSum += bAddrU;
                                bBuffer[iBufferIndx++] = bAddrH; iSum += bAddrH;
                                bBuffer[iBufferIndx++] = bAddrL; iSum += bAddrL;
                                bBuffer[iBufferIndx++] = cCmdErasePage; iSum += cCmdErasePage;
                                bBuffer[iBufferIndx++] = bSize; iSum += bSize;

                                // Calculate checksum and add to buffer
                                iSum %= 256;
                                iChecksum = Convert.ToInt16((256 - iSum) % 256);
                                bBuffer[iBufferIndx++] = Convert.ToByte(iChecksum);

                                // Send address, command, packetsize & checksum
                                pobjPort.SendBytes(ref bBuffer, iBufferIndx);
                                
                                // Get response, checksum ok
                                ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                                if ( bProcessWriteResponseResult == false ) {
                                    return false;
                                } else if ( debugMode == true && bRetry == false ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                                }

                                //
                                Application.DoEvents();
                            } while (bRetry == true && bAbort == false);
                        }//if ( pobjDevice.pageSizeR > 1 ) {


                        //---------------------------------------------------------
                        // Iterate rows, write all rows in page even if not in hex-file
                        //---------------------------------------------------------
                        for ( iRow = iPage * pobjDevice.pageSizeR; iRow < iPage * pobjDevice.pageSizeR + pobjDevice.pageSizeR; iRow++ ) {
                            iRetries = 0;
                            do {
                                if ( clsds30Loader.debugMode ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing row " + iRow.ToString() + "...", iTabLevel) );
                                }

                                iBufferIndx = 0;
                                iSum = 0;

                                // Calculate
                                iAddrP = iRow * (pobjDevice.rowsizeW * 2);				//24-bit address in memory to write to
                                bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
                                bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
                                bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) >> 0 );	//address low byte
                                bSize = Convert.ToByte( pobjDevice.rowsizeW * 3 + 1/*checksum*/ );

                                //
                                bBuffer[iBufferIndx++] = bAddrU; iSum += bAddrU;
                                bBuffer[iBufferIndx++] = bAddrH; iSum += bAddrH;
                                bBuffer[iBufferIndx++] = bAddrL; iSum += bAddrL;
                                bBuffer[iBufferIndx++] = cCmdWriteRow; iSum += cCmdWriteRow;
                                bBuffer[iBufferIndx++] = bSize; iSum += bSize;

                                // Calculate startindex in hex-buffer
                                iAddrP = iRow * pobjDevice.rowsizeW * 3;

						        // Buffer row, upper-low-high, ordered in ParseHex()
						        for ( int iByte = 0; iByte < pobjDevice.rowsizeW * 3; iByte++ ) {
								    bProg = Convert.ToByte( pobjHex.iProgMem[iAddrP + iByte] );
							        bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;
						        }

                                // Calculate checksum and add to buffer
                                iSum %= 256;
                                iChecksum = Convert.ToInt16((256 - iSum) % 256);
                                bBuffer[iBufferIndx++] = Convert.ToByte(iChecksum);

                                // Send address, command, packetsize, row & checksum
                                pobjPort.SendBytes(ref bBuffer, iBufferIndx);

                                // Get response, checksum ok
                                ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                                if ( bProcessWriteResponseResult == false ) {
                                    return false;
                                } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                                }

                                //
                                Application.DoEvents();
                            } while ( bRetry == true && bAbort == false );

                            //
                            if (bAbort == true) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
                                return false;
                            }

                            //
                            iBytesWritten += pobjDevice.rowsizeW * 3;
                            OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );

                        }//for ( iRow = 0; iRow < iRowUsedBufferSize; iRow++ ) {

                    } //if ( bPageUsed[iIter] == true ) {

                }//for ( iPage = 0; iPage < iPageUsedBufferSize; iPage++ ) {

                if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteProgram == true ) {


            //--------------------------------------------------------------------------
            // Write Eeprom
            //--------------------------------------------------------------------------
            if ( pbWriteEeprom == true ) {
			    if ( clsds30Loader.debugMode == false ) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom...", iTabLevel) );
                }
				
			    for ( int iEepromWord = 0; iEepromWord < pobjDevice.EepromSizeB/2; iEepromWord++ ) {					
				    if ( pobjHex.bEEWordUsed[iEepromWord] == true ) {
						iRetries = 0;
						
					    do {
                            if ( clsds30Loader.debugMode ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing Eeprom word " + iEepromWord.ToString() + " (0x" + ((pobjHex.iEEMem[iEepromWord*2+1]<<8)+pobjHex.iEEMem[iEepromWord * 2]).ToString("X4") + ")...", iTabLevel) );
                            }

						    iBufferIndx = 0;
						    iSum = 0;
    		                
		                    // Calculate
		                    iAddrP = pobjDevice.EepromStartAddress + iEepromWord * 2;		//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
						    bSize = Convert.ToByte( 2 + 1/*checksum*/ );
    						
						    //
						    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
						    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
						    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
                            bBuffer[ iBufferIndx++ ] = cCmdWriteEEWord; iSum += cCmdWriteEEWord;
						    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;

							// Low byte
						    bProg = Convert.ToByte( pobjHex.iEEMem[iEepromWord*2 + 0] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum += bProg;
							
							// High byte								    
						    bProg = Convert.ToByte( pobjHex.iEEMem[iEepromWord*2 + 1] );
						    bBuffer[ iBufferIndx++ ] = bProg; iSum += bProg;
    		                
						    // Calculate checksum and add to buffer
						    iSum %= 256;
						    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
						    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
    						
						    // Send address, size, word and checksum
						    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
    		                
		                    // Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "Eeprom", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                            }
    		                
		                    //
						    Application.DoEvents();
					    } while ( bRetry == true && bAbort == false );
    		            
		                //
					    if ( bAbort == true ) {
						    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
						    return false;
					    }
    		            
    		            //
					    iBytesWritten += 2;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
				     }//if (  bEEWordUsed[iEepromWord] == true ) {				     
				     
			    }//for ( iIter = 0; iIter < pobjDevice.EepromSizeB/2; iIter++ ) {
			    
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                }
            }//if ( pbWriteEeprom == true ) {
            

            //--------------------------------------------------------------------------
			// Write Config Bits
			//--------------------------------------------------------------------------
            if ( pbWriteConfigs == true && pobjDevice.family.name == "dsPIC30F" ) {
				if ( clsds30Loader.debugMode == false) {
                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing configs...", iTabLevel) );
                }
								
				for ( int iConfig = 0;  iConfig < clsHex30F.iConfigWordsUsedBufferSize; iConfig++ ) {
					if ( pobjHex.bConfigWordUsed[iConfig] == true ) {
						iRetries = 0;
						
						do {
                            if ( clsds30Loader.debugMode == true ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing config word " + iConfig.ToString() + " (0x" + ((pobjHex.iConfigMem[iConfig * 2+1]<<8)+pobjHex.iConfigMem[iConfig * 2]).ToString("X4") + ")...", iTabLevel) );
                            }

							iBufferIndx = 0;
							iSum = 0;						
						
							// Calculate
							iAddrP = 0xF80000 + iConfig * 2;						//24-bit address in memory to write to
						    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
						    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
						    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte                            
							bSize = Convert.ToByte( 2 + 1/*checksum*/ );
							
							//
							bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
							bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
							bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
                            bBuffer[ iBufferIndx++ ] = cCmdWriteConfig; iSum += cCmdWriteConfig;
							bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;	
		                    
							// Buffer config word
							for ( int iIter = 0; iIter < 2; iIter++ ) {
								bProg = Convert.ToByte( pobjHex.iConfigMem[ iConfig * 2 + iIter ] );
								bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;	
							}
							
							// Calculate checksum and add to buffer
							iSum %= 256;
							iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
							bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
							
							// Send row+checksum
							pobjPort.SendBytes( ref bBuffer, iBufferIndx );
		                    
							// Get response, checksum ok
                            ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                            if ( bProcessWriteResponseResult == false ) {
                                return false;
                            } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                            }

                            //
                            Application.DoEvents();
						} while ( bRetry == true && bAbort == false );
		                
						//
						if ( bAbort == true ) {
							OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
							return false;
						}
						
                        //
						iBytesWritten += 2;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
					}//if ( bConfigWordUsed[iIter] == true ) {
					
				}//for ( iIter = 0;  iIter < iConfigUsedBufferSize; iIter++ ) { 
				
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                }
			}//if ( bWriteConfigs == true )


            //--------------------------------------------------------------------------
			// Write Config Bits
			//--------------------------------------------------------------------------
            if ( pbWriteConfigs == true && (pobjDevice.family.name == "dsPIC33FJ" || pobjDevice.family.name == "PIC24H") ) {
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing configs...", iTabLevel) );
                }
				
				for ( int iConfig = 0; iConfig < clsHex33FJ.iConfigWordsUsedBufferSize; iConfig++ ) {
					if ( pobjHex.bConfigWordUsed[iConfig] == true ) {
                        iRetries = 0;
						do {
                            for ( int iByte = 0; iByte < 2; iByte++ ) {
                                if ( clsds30Loader.debugMode ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.info, "Writing config " + iConfig.ToString() + " byte " + iByte.ToString() + "...", iTabLevel) );
                                }

                                iBufferIndx = 0;
							    iSum = 0;						
    						
							    // Calculate
							    iAddrP = 0xF80000 + iConfig * 2 + iByte;				//24-bit address in memory to write to
							    bAddrU = Convert.ToByte( (iAddrP & 0xff0000) >> 16 );	//address upper byte			               
							    bAddrH = Convert.ToByte( (iAddrP & 0x00ff00) >> 8 );	//address high byte
							    bAddrL = Convert.ToByte( (iAddrP & 0x0000ff) );			//address low byte
							    bSize = Convert.ToByte( 1 + 1/*checksum*/ );
    							
							    //
							    bBuffer[ iBufferIndx++ ] = bAddrU;	iSum += bAddrU;
							    bBuffer[ iBufferIndx++ ] = bAddrH;	iSum += bAddrH;
							    bBuffer[ iBufferIndx++ ] = bAddrL;	iSum += bAddrL;
							    bBuffer[ iBufferIndx++ ] = cCmdWriteConfig;iSum += cCmdWriteConfig;
							    bBuffer[ iBufferIndx++ ] = bSize;	iSum += bSize;	
    		                    
							    // Buffer config word
							    bProg = Convert.ToByte( pobjHex.iConfigMem[ iConfig * 2 + iByte ] );
							    bBuffer[ iBufferIndx++ ] = bProg; iSum = (iSum + bProg) % 256;	
    							
							    // Calculate checksum and add to buffer
							    iSum %= 256;
							    iChecksum = Convert.ToInt16( (256 - iSum) % 256 );
							    bBuffer[ iBufferIndx++ ] = Convert.ToByte(iChecksum);
    							
							    // Send row+checksum
							    pobjPort.SendBytes( ref bBuffer, iBufferIndx );
    		                    
							    // Get response, checksum ok
                                ProcessWriteResponse( pobjPort, ref iRetries, iTabLevel, "flash", ref bRetry, ref bProcessWriteResponseResult );
                                if ( bProcessWriteResponseResult == false ) {
                                    return false;
                                } else if ( clsds30Loader.debugMode == true && bRetry == false ) {
                                    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );
                                }
                            }
                            Application.DoEvents();
						} while ( bRetry == true && bAbort == false );
		                
						//
						if ( bAbort == true ) {
							OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.error, "aborted by user", -1) );
							return false;
						}
						
                        //
						iBytesWritten += 2;
                        OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.progress, "", (100 * iBytesWritten) / iByteToWrite ) );
					}//if ( bConfigUsed[iIter] == true ) {
					
				}//for ( iIter = 0;  iIter < iConfigUsedBufferSize; iIter++ ) { 
				
                if ( clsds30Loader.debugMode == false ) {
				    OnDownloading( new clsDownloadingEventArgs(clsDownloadingEventArgs.EventType.success, "ok", -1) );				
                }
			}//if ( bWriteConfigs == true )


            //--------------------------------------------------------------------------
            // Write completed
            //--------------------------------------------------------------------------
            return true;
        }// WriteFlash16()






        #endregion

    }// Class: clsds30Loader
}
